import uuid

from django.test import TestCase, override_settings

from arkindex.dataimport.utils import get_default_farm_id
from ponos.models import Farm

DEFAULT_FARM_ID = str(uuid.uuid4())


class TestDataImportUtils(TestCase):
    """
    Test utilities functions for dataimport module
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.corn_farm = Farm.objects.create(id=DEFAULT_FARM_ID, name='Corn farm')
        cls.barley_farm = Farm.objects.create(name='Barley farm')

    def setUp(self):
        # Force clean the module default farm cached global variable
        from arkindex.dataimport import utils
        setattr(utils, '__default_farm_id', None)

    @override_settings(PONOS_DEFAULT_FARM=DEFAULT_FARM_ID)
    def test_config_ponos_farm_setting(self):
        """
        Default farm may be defined in settings
        """
        with self.assertNumQueries(1):
            self.assertEqual(str(get_default_farm_id()), str(self.corn_farm.id))

    def test_default_ponos_farm(self):
        """
        In case no default farm is defined, a random farm is selected as default.
        Subsequent calls does not hit the database
        """
        farm_id = None
        with self.assertNumQueries(1):
            farm_id = get_default_farm_id()
        self.assertIn(farm_id, Farm.objects.values_list('id', flat=True))
        with self.assertNumQueries(0):
            self.assertEqual(get_default_farm_id(), farm_id)

    def test_default_farm_no_existing_farm(self):
        """
        In case no farm exists, a default one is created
        """
        Farm.objects.all().delete()
        self.assertEqual(Farm.objects.count(), 0)
        farm_id = None
        with self.assertNumQueries(2):
            farm_id = get_default_farm_id()
            self.assertIsInstance(farm_id, uuid.UUID)
        self.assertTrue(Farm.objects.filter(id=farm_id).exists())
        with self.assertNumQueries(0):
            self.assertEqual(get_default_farm_id(), farm_id)
