import math
from textwrap import dedent

from django.conf import settings
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.exceptions import NotFound, ValidationError
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from rest_framework.views import APIView
from SolrClient import SolrClient
from SolrClient.exceptions import SolrError

from arkindex.documents.serializers.search import CorpusSearchQuerySerializer, CorpusSearchResultSerializer
from arkindex.project.mixins import CorpusACLMixin

solr = SolrClient(settings.SOLR_API_URL)


@extend_schema_view(
    get=extend_schema(
        operation_id='SearchCorpus',
        tags=['search'],
        request=CorpusSearchQuerySerializer,
        parameters=[
            OpenApiParameter(
                'query',
                type=str,
                description=dedent('''
                Search query on transcription text, metadata value or element and entity name.

                * By default, will search for the exact word (`a` will not match `zab`)
                * Prefix and suffix the word with `*` to search it in a text (`*a*` will match `zab`)
                * By default, will look for only one of each word (`a b` is translated to `a or b`).
                * Combine words with `and` to require multiple words.
                * Prefix a word with `-` or `!` to filter out items that contain a word.
                * Add `*` around multiple words to require them all to be present
                * in a specific order (`*a*b*`).
                * Prefix parentheses with a `-` or `!` to require multiple words to be missing (`-(a b)`).
                * Add quotes around multiple words to search a specific sentence (`"a b"`).
                '''),
                required=True,
            ),
            OpenApiParameter(
                'sources',
                type=dict,
                description='List of sources to be searched among "element", "transcription", "metadata", "entity"',
                default="{'element', 'transcription', 'metadata', 'entity'}",
                required=False
            ),
            OpenApiParameter(
                'facets',
                type=dict,
                description=dedent('''
                Dictionary (key, value) of facets to filter. The key can be:

                * element_type
                * element_worker
                * transcription_worker
                * classification_name
                * classification_worker
                * metadata_name
                * metadata_type
                * metadata_worker
                * entity_type
                * entity_worker
                '''),
                default={},
                required=False
            ),
            OpenApiParameter(
                'only_facets',
                type=bool,
                description='Returns only the facets. Results will be none',
                default=False,
                required=False
            )
        ],
        responses=CorpusSearchResultSerializer
    )
)
class CorpusSearch(CorpusACLMixin, APIView):
    """
    Get a list of Solr documents in a specific corpus whose transcription text, metadata value or element and entity name correspond to the given query.
    """

    def build_solr_query(self, query, sources, facets):
        sources_query = " OR ".join([f'{source}_text:({query})' for source in sources])
        facets_query = " AND ".join([f'{key}:"{value}"' for key, value in facets.items()])
        return f'({sources_query}) AND ({facets_query})' if facets_query else sources_query

    def solr_search(self, collection_name, query, page):
        return solr.query(collection_name, {
            'q': query,
            'start': (page - 1) * settings.SOLR_PAGINATION_SIZE,
            'rows': settings.SOLR_PAGINATION_SIZE,
            'facet': True,
            'facet.field': [
                'element_type',
                'element_worker',
                'transcription_worker',
                'classification_name',
                'classification_worker',
                'metadata_name',
                'metadata_type',
                'metadata_worker',
                'entity_type',
                'entity_worker'
            ]
        })

    def build_pagination(self, url, page, nb_results):
        nb_page = math.ceil(nb_results / settings.SOLR_PAGINATION_SIZE)
        previous_url = replace_query_param(url, 'page', page - 1) if page > 1 else None
        next_url = replace_query_param(url, 'page', page + 1) if page < nb_page else None
        return previous_url, next_url

    def get(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['search']:
            raise ValidationError(['Search features are not available on this instance.'])

        corpus = self.get_corpus(kwargs['pk'])
        collection_name = f'project-{corpus.id}'

        if not corpus.indexable:
            raise ValidationError([f'Corpus {corpus.id} is not indexable.'])

        if not solr.collections.exists(collection_name):
            raise NotFound(f'Corpus index {collection_name} not found.')

        # Serialize params
        params = request.query_params.dict()
        # Convert the sources string into a valid list
        # Support both possible urls
        if 'sources' in params or 'sources[]' in params:
            params['sources'] = request.query_params.getlist('sources') + request.query_params.getlist('sources[]')
        serializer = CorpusSearchQuerySerializer(data=params)
        serializer.is_valid(raise_exception=True)
        q, sources, only_facets, page = map(
            lambda key: serializer.validated_data.pop(key),
            ('query', 'sources', 'only_facets', 'page')
        )

        # Paginated Solr search
        query = self.build_solr_query(q, sources, serializer.validated_data)
        try:
            results = self.solr_search(collection_name, query, page)
        except SolrError:
            raise ValidationError([f'Query "{q}" is not valid.'])

        # Build response
        nb_results = results.get_num_found()
        previous_url, next_url = self.build_pagination(request.build_absolute_uri(), page, nb_results)

        return Response(CorpusSearchResultSerializer({
            'count': nb_results if not only_facets else 0,
            'number': page,
            'next': next_url,
            'previous': previous_url,
            'results': results.docs if not only_facets else None,
            'facets': results.get_facets()
        }).data, status=status.HTTP_200_OK)
