#!/usr/bin/env python3
from uuid import uuid4

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Q
from django.db.utils import IntegrityError
from rest_framework.authtoken.models import Token

from arkindex.dataimport.models import (
    Repository,
    RepositoryType,
    Worker,
    WorkerVersion,
    WorkerVersionGPUUsage,
    WorkerVersionState,
)
from arkindex.images.models import ImageServer
from arkindex.users.models import User
from ponos.models import Farm

# Constants used in architecture project
IMAGE_SERVER_ID = 12345
IMAGE_SERVER_BUCKET = 'iiif'
IMAGE_SERVER_REGION = 'local'
PONOS_FARM_ID = "001e411a-1111-2222-3333-444455556666"
PONOS_FARM_NAME = 'Bootstrap farm'
PONOS_FARM_SEED = "b12868101dab84984481741663d809d2393784894d6e807ceee0bd95051bf971"
IMPORT_WORKER_VERSION_ID = 'f2bb8dd7-55e9-49ae-9bd9-b1d2e5d491b9'
IMPORT_WORKER_SLUG = 'file_import'
IMPORT_WORKER_REPO = 'https://gitlab.com/arkindex/tasks'
IMPORT_WORKER_REVISION_MESSAGE = 'File import worker bootstrap'
IMPORT_WORKER_REVISION_AUTHOR = 'Dev Bootstrap'
INTERNAL_API_TOKEN = "deadbeefTestToken"


class Command(BaseCommand):
    help = 'Build required Database model instances for local development.'

    def success(self, msg):
        """Helper to display success messages"""
        self.stdout.write(self.style.SUCCESS(f"✅ {msg}"))

    def warn(self, msg):
        """Helper to display warning messages"""
        self.stdout.write(self.style.WARNING(f"⚠ {msg}"))

    def fail(self, msg):
        """Helper to display error messages"""
        self.stdout.write(self.style.ERROR(f"❌ {msg}"))

    def check_user(self, user):
        """Ensure a user is admin + internal"""
        if user.is_internal and user.is_admin:
            self.success(f"Internal user {user} is valid")
        else:
            user.is_internal = True
            user.is_admin = True
            user.save()
            self.warn(f"Updated user {user} to internal+admin")

    def handle(self, **options):
        # Never allow running this script in production
        if not settings.DEBUG:
            self.fail("You cannot run this script in production.")
            return

        # a Ponos farm with specific ID & seed
        try:
            farm = Farm.objects.get(id=PONOS_FARM_ID)
            update_farm = False

            if farm.seed != PONOS_FARM_SEED:
                farm.seed = PONOS_FARM_SEED
                update_farm = True

            if farm.name != PONOS_FARM_NAME:
                farm.name = PONOS_FARM_NAME
                update_farm = True

            if update_farm:
                farm.save()
                self.warn(f"Ponos farm {farm.name} updated")
            else:
                self.success("Ponos farm valid")
        except Farm.DoesNotExist:
            Farm.objects.create(
                id=PONOS_FARM_ID,
                name=PONOS_FARM_NAME,
                seed=PONOS_FARM_SEED,
            )
            self.success("Ponos farm created")

        # an internal API user with a specific token
        try:
            token = Token.objects.get(key=INTERNAL_API_TOKEN)
            self.check_user(token.user)
        except Token.DoesNotExist:
            # Create a new internal user
            user, _ = User.objects.get_or_create(
                email='internal+bootstrap@teklia.com',
                defaults={
                    'display_name': 'Bootstrap Internal user',
                    'is_internal': True,
                    'is_admin': True,
                }
            )
            self.success("Created internal user")
            self.check_user(user)

            # Finally create a specific token for that user
            if hasattr(user, "auth_token"):
                # Support One-To-One relation
                user.auth_token.delete()
            Token.objects.create(key=INTERNAL_API_TOKEN, user=user)
            self.success(f"Created token {INTERNAL_API_TOKEN}")

        # an image server for local cantaloupe https://ark.localhost/iiif/2
        try:
            server = ImageServer.objects.get(url='https://ark.localhost/iiif/2')
            if server.id != IMAGE_SERVER_ID:
                # Migrate existing images & server id in a single transaction
                with transaction.atomic():
                    server.images.update(server_id=IMAGE_SERVER_ID)
                    ImageServer.objects.filter(id=server.id).update(id=IMAGE_SERVER_ID)
                self.warn(f"Image server {server.id} updated to {IMAGE_SERVER_ID}")

                # Update internal reference for updates below
                server.id = IMAGE_SERVER_ID

            # Update base settings
            if server.s3_bucket != IMAGE_SERVER_BUCKET or server.s3_region != IMAGE_SERVER_REGION or not server.validated:
                server.s3_bucket = IMAGE_SERVER_BUCKET
                server.s3_region = IMAGE_SERVER_REGION
                server.validated = True
                server.save()
                self.warn("Updated image server S3 settings")
            else:
                self.success(f"Image server {server.id} valid")
        except ImageServer.DoesNotExist:
            try:
                server = ImageServer.objects.create(
                    id=IMAGE_SERVER_ID,
                    url='https://ark.localhost/iiif/2',
                    s3_bucket=IMAGE_SERVER_BUCKET,
                    s3_region=IMAGE_SERVER_REGION,
                    display_name='Development local server',
                    validated=True,
                )
                self.success("Image server created")
            except IntegrityError as e:
                self.fail(f"Failed to create image server: {e}")
                return

        # Check there is not already a local server with invalid path
        # We'll merge its image into the new one
        # This bad server may have been created by automatic IIIF server detection
        try:
            bad_server = ImageServer.objects.get(url='https://ark.localhost/iiif')
            bad_server.merge_into(server)
            self.warn(f"Merged images from {bad_server.id} into {server.id}")

            bad_server.delete()
            self.warn("Deleted old server")
        except ImageServer.DoesNotExist:
            pass

        # Create a fake worker version on a fake worker on a fake repo with a fake revision for file imports
        repo, created = Repository.objects.get_or_create(
            url=IMPORT_WORKER_REPO,
            type=RepositoryType.Worker,
            defaults={
                "hook_token": str(uuid4()),
            }
        )
        if created:
            self.success(f'Created Git repository for {IMPORT_WORKER_REPO}')
        else:
            self.success(f'Repository for {IMPORT_WORKER_REPO} exists')

        worker, created = Worker.objects.update_or_create(
            repository=repo,
            slug=IMPORT_WORKER_SLUG,
            defaults={'name': 'File import', 'type': 'import'},
        )
        if created:
            self.success(f'Created {IMPORT_WORKER_SLUG} worker')
        else:
            self.success(f'Updated {IMPORT_WORKER_SLUG} worker')

        revision = repo.revisions.filter(
            Q(message=IMPORT_WORKER_REVISION_MESSAGE)
            | Q(author=IMPORT_WORKER_REVISION_AUTHOR)
        ).first()
        if revision:
            revision.message = IMPORT_WORKER_REVISION_MESSAGE
            revision.author = IMPORT_WORKER_REVISION_AUTHOR
            self.success(f'Updated revision {revision.hash}')
        else:
            revision = repo.revisions.create(
                hash=uuid4().hex,
                message=IMPORT_WORKER_REVISION_MESSAGE,
                author=IMPORT_WORKER_REVISION_AUTHOR,
            )
            self.success(f'Created revision {revision.hash}')

        version, created = worker.versions.get_or_create(
            revision=revision,
            defaults={
                'id': IMPORT_WORKER_VERSION_ID,
                'configuration': {},
                'state': WorkerVersionState.Created,
                'gpu_usage': WorkerVersionGPUUsage.Disabled,
                'docker_image': None,
                'docker_image_iid': None,
            }
        )
        if created:
            self.success(f'Created worker version {version.slug}')
        else:
            # The version exists; ensure it has the right UUID
            if str(version.id) != IMPORT_WORKER_VERSION_ID:
                # If there is a version with the right ID, just nuke it
                bad_version = WorkerVersion.objects.filter(id=IMPORT_WORKER_VERSION_ID).first()
                if bad_version:
                    bad_version.delete()
                    self.warn(f'Deleted invalid worker version {bad_version.slug}')

                with transaction.atomic():
                    version.elements.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.transcriptions.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.transcription_entities.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.classifications.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.entities.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.metadatas.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.worker_runs.update(version_id=IMPORT_WORKER_VERSION_ID)
                    version.activities.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    version.corpus_cache.update(worker_version_id=IMPORT_WORKER_VERSION_ID)
                    WorkerVersion.objects.filter(id=version.id).update(id=IMPORT_WORKER_VERSION_ID)
                version.id = IMPORT_WORKER_VERSION_ID
                self.success(f'Updated worker version {version.slug} to ID {version.id}')

            # Ensure it has the right attributes
            version.configuration = {}
            version.state = WorkerVersionState.Created
            version.gpu_usage = WorkerVersionGPUUsage.Disabled
            version.docker_image = None
            version.docker_image_iid = None
            version.save()

            self.success(f'Updated attributes of worker version {version.slug}')
