#!/usr/bin/env python3
import time
from urllib.parse import urljoin

import requests
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import Count

from arkindex.images.models import Image, ImageServer
from arkindex.project.aws import S3FileStatus


class Command(BaseCommand):
    help = 'Display statistics for a Telegraf agent'

    def add_arguments(self, parser):
        parser.add_argument(
            '--influxdb',
            help='Post statistics to the InfluxDB API at a given URL',
            nargs='?',
            const=settings.INFLUXDB_API_URL,
        )

    def handle(self, *args, **options):
        self.options = options

        # Image server statistics, in three SQL queries
        checked_counts = dict(
            Image.objects
                 .filter(status=S3FileStatus.Checked)
                 .values_list('server')
                 .annotate(Count('server'))
        )
        error_counts = dict(
            Image.objects
                 .filter(status=S3FileStatus.Error)
                 .values_list('server')
                 .annotate(Count('server'))
        )
        for server in ImageServer.objects.annotate(img_total=Count('images')):
            self.output(
                'images',
                tags={'imageserver': server.grafana_tag},
                total=server.img_total,
                checked=checked_counts.get(server.id, 0),
                errors=error_counts.get(server.id, 0),
            )

    def dict_to_str(self, data):
        return ','.join([
            '{}={}'.format(k, str(v).replace(' ', '_'))
            for (k, v) in data.items()
        ])

    def output(self, name, tags={}, **values):
        '''
        Output on stdout in Influx line protocol
        https://docs.influxdata.com/influxdb/v1.6/write_protocols/line_protocol_tutorial/
        '''
        timestamp = time.time() * 10 ** 9  # in nano seconds
        tags_str = ',' + self.dict_to_str(tags) if tags else ''
        values_str = self.dict_to_str(values)
        line = '{}{} {} {:.0f}'.format(name, tags_str, values_str, timestamp)
        self.stdout.write(line)

        if not self.options.get('influxdb'):
            return
        resp = requests.post(urljoin(self.options['influxdb'], 'write'), params={'db': 'arkindex'}, data=line)
        resp.raise_for_status()
