from rest_framework import serializers

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Entity, EntityLink, EntityRole, EntityType, TranscriptionEntity
from arkindex.documents.serializers.light import CorpusLightSerializer, InterpretedDateSerializer
from arkindex.project.serializer_fields import EnumField


class BaseEntitySerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)
    dates = InterpretedDateSerializer(many=True, source='get_dates', read_only=True)
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    worker_version_id = serializers.PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'dates',
            'worker_version_id',
        )
        read_only_fields = (
            'id',
            'dates',
            'worker_version_id',
        )


class EntityRoleSerializer(serializers.ModelSerializer):
    """
    Serialize a role between two types of entity
    """
    parent_type = EnumField(EntityType)
    child_type = EnumField(EntityType)

    class Meta:
        model = EntityRole
        fields = (
            'id',
            'parent_name',
            'child_name',
            'parent_type',
            'child_type'
        )

    def validate(self, data):
        assert 'corpus' not in data
        assert self.context.get('request') is not None
        corpus_id = self.context['request'].parser_context['kwargs']['pk']
        corpus = Corpus.objects.writable(self.context['request'].user).filter(id=corpus_id).first()
        if corpus is None:
            raise serializers.ValidationError({
                'corpus': ['You do not have write access to this corpus'],
                'id': corpus_id,
            })
        data['corpus'] = corpus
        return data


class CreateEntityRoleErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="The corpus ID.")
    corpus = serializers.ListField(child=serializers.CharField(), required=False, help_text="Errors that occured during corpus ID field validation.")


class EntityLinkSerializer(serializers.ModelSerializer):
    """
    Serialize an entity link with its child, parent and role
    """
    parent = BaseEntitySerializer()
    child = BaseEntitySerializer()
    role = EntityRoleSerializer()

    class Meta:
        model = EntityLink
        fields = (
            'id',
            'parent',
            'child',
            'role'
        )


class EntitySerializer(BaseEntitySerializer):
    """
    Serialize an entity with its metadata
    """
    corpus = CorpusLightSerializer(read_only=True)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)

    class Meta:
        model = Entity
        fields = BaseEntitySerializer.Meta.fields + (
            'corpus',
            'children',
            'parents'
        )
        read_only_fields = BaseEntitySerializer.Meta.read_only_fields = (
            'corpus',
            'children',
            'parents',
        )

    def validate(self, data):
        if data.get('validated'):
            data['moderator'] = self.context['request'].user
        return data


class EntityCreateSerializer(BaseEntitySerializer):
    """
    Serialize an entity with a possible parents and children
    """
    corpus = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        style={'base_template': 'input.html'},
    )
    metas = serializers.HStoreField(child=serializers.CharField(), required=False)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)
    worker_version = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.all(),
        default=None,
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'corpus',
            'parents',
            'children',
            'worker_version'
        )
        read_only_fields = (
            'id',
            'children',
            'parents',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora


class EntityLinkCreateSerializer(EntityLinkSerializer):
    """
    Serialize an entity with a possible parents and children
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.none(),
        style={'base_template': 'input.html'},
    )
    child = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.none(),
        style={'base_template': 'input.html'},
    )
    role = serializers.PrimaryKeyRelatedField(
        queryset=EntityRole.objects.none(),
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = EntityLink
        fields = EntityLinkSerializer.Meta.fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        entities = Entity.objects.all().filter(corpus__in=corpora)
        roles = EntityRole.objects.all().filter(corpus__in=corpora)
        self.fields['parent'].queryset = entities
        self.fields['child'].queryset = entities
        self.fields['role'].queryset = roles

    def validate(self, data):
        data = super().validate(data)
        link = EntityLink(**data)
        link.full_clean()
        return data


class TranscriptionEntitySerializer(serializers.ModelSerializer):
    """
    Serialise the link between an entity and a transcription
    Transcription element must be passed in serializer context
    """
    offset = serializers.IntegerField(min_value=0)
    length = serializers.IntegerField(min_value=1)
    worker_version_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.all(),
        source='worker_version',
        default=None,
        style={'base_template': 'input.html'},
    )
    confidence = serializers.FloatField(min_value=0, max_value=1, default=None)

    class Meta:
        model = TranscriptionEntity
        fields = ('entity', 'offset', 'length', 'worker_version_id', 'confidence')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        self.fields['entity'].queryset = Entity.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user)
        )

    def validate(self, data):
        data = super().validate(data)
        errors = {}

        data['transcription'] = self.context['transcription']
        if not data['transcription']:
            errors['transcription'] = ['invalid UUID or Corpus write-access is forbidden']

        user = self.context['request'].user
        if data.get('worker_version') and not (user.is_authenticated and user.is_internal):
            errors['worker_version_id'] = [
                'An internal user is required to link an entity to a transcription with a worker version.'
            ]

        if errors:
            raise serializers.ValidationError(errors)

        return data


class TranscriptionEntityDetailsSerializer(TranscriptionEntitySerializer):
    """
    Serialize the link between a transcription and a detailled entity
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['entity'] = BaseEntitySerializer()
