from datetime import datetime, timedelta, timezone
from io import StringIO
from textwrap import dedent
from unittest.mock import call, patch

from botocore.exceptions import ClientError
from django.core.management import call_command
from django.test import override_settings

from arkindex.dataimport.models import DataFile
from arkindex.documents.models import CorpusExport, CorpusExportState
from arkindex.project.tests import FixtureTestCase


@override_settings(AWS_EXPORT_BUCKET='export')
class TestCleanupCommand(FixtureTestCase):

    def cleanup(self, args=[]):
        """
        Run the cleanup command, capturing its output
        """
        output = StringIO()
        call_command('cleanup', args + ['--no-color'], stdout=output, stderr=output)
        return output.getvalue().strip()

    @patch('arkindex.project.aws.s3.Object')
    def test_cleanup(self, s3_object_mock):
        young_export = self.corpus.exports.create(user=self.superuser)
        # Use a fake creation time to make old exports
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            self.corpus.exports.create(user=self.user, state=CorpusExportState.Failed)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Removing 2 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )

        self.assertEqual(CorpusExport.objects.count(), 1)
        self.assertEqual(CorpusExport.objects.get(), young_export)
        self.assertEqual(s3_object_mock.call_count, 1)
        self.assertEqual(s3_object_mock.call_args, call('export', str(done_export.id)))
        self.assertEqual(s3_object_mock.return_value.delete.call_count, 1)

    def test_nothing(self):
        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )

    @patch('arkindex.project.aws.s3.Object')
    def test_s3_not_found(self, s3_object_mock):
        s3_object_mock.return_value.delete.side_effect = ClientError({'Error': {'Code': '404'}}, 'delete_object')

        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Export {done_export.id} not found on S3, skipping
                Removing 1 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )

        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(s3_object_mock.call_count, 1)
        self.assertEqual(s3_object_mock.call_args, call('export', str(done_export.id)))
        # s3.Object.delete should only be called once, not retried
        self.assertEqual(s3_object_mock.return_value.delete.call_count, 1)

    @patch('arkindex.project.aws.s3.Object')
    def test_s3_error(self, s3_object_mock):
        error = ClientError({'Error': {'Code': '500'}}, 'delete_object')
        # Fail twice, then delete successfully
        s3_object_mock.return_value.delete.side_effect = [error, error, None]

        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Removing 1 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )

        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(s3_object_mock.call_count, 1)
        self.assertEqual(s3_object_mock.call_args, call('export', str(done_export.id)))
        self.assertEqual(s3_object_mock.return_value.delete.call_count, 3)

    @patch('arkindex.project.aws.s3')
    def test_cleanup_trashed_datafiles(self, s3_mock):
        DataFile.objects.bulk_create([
            DataFile(
                name=f'test{i}.txt',
                size=42,
                content_type='text/plain',
                corpus=self.corpus,
                trashed=True if i % 2 else False
            ) for i in range(0, 2)
        ])

        trashed_df = DataFile.objects.filter(trashed=True).get()
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 1 DataFiles marked as trashed from S3 and the database…
                Deleting DataFile {trashed_df.id}…
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )
        self.assertEqual(DataFile.objects.count(), 1)
        self.assertEqual(DataFile.objects.filter(trashed=True).count(), 0)
        self.assertEqual(s3_mock.Object().delete.call_count, 1)

    @patch('arkindex.project.aws.s3')
    def test_cleanup_trashed_datafiles_ignore_s3_errors(self, s3_mock):
        """
        Test the cleanup command tries multiple times to delete from S3, but ignores errors if it fails
        """
        DataFile.objects.bulk_create([
            DataFile(
                name=f'test{i}.txt',
                size=42,
                content_type='text/plain',
                corpus=self.corpus,
                trashed=True if i % 2 else False
            ) for i in range(0, 2)
        ])

        exceptions = [
            ClientError({'Error': {'Code': '500'}}, 'delete_object'),
            ClientError({'Error': {'Code': '500'}}, 'delete_object'),
            ValueError('Something went wrong'),
        ]

        def _raise(*args, **kwargs):
            raise exceptions.pop(0)

        s3_mock.Object().delete.side_effect = _raise

        trashed_df = DataFile.objects.filter(trashed=True).get()
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Deleting 1 DataFiles marked as trashed from S3 and the database…
                Deleting DataFile {trashed_df.id}…
                Could not delete DataFile {trashed_df.id} from S3: Something went wrong
                Successfully cleaned up DataFiles marked as trashed.
                """
            ).strip()
        )
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(DataFile.objects.filter(trashed=True).count(), 1)
        self.assertEqual(s3_mock.Object().delete.call_count, 3)
