import json
import sqlite3
import tempfile
from pathlib import Path
from unittest.mock import patch

from django.contrib.gis.geos.linestring import LinearRing
from django.core.management import CommandError, call_command

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.export import export_corpus
from arkindex.documents.models import Corpus, Element, EntityType, Transcription
from arkindex.documents.tasks import corpus_delete
from arkindex.images.models import Image, ImageServer
from arkindex.project.tests import FixtureTestCase

BASE_DIR = Path(__file__).absolute().parent


class TestLoadExport(FixtureTestCase):

    def clean_dump_data(self, path, corpus=None):
        # Some fields like id, created and updated are not comparable.
        # We keep only the comparable fields for each imported model.

        unexpected_fields_by_model = {
            'documents.elementtype': ['display_name', 'indexable'],
            'documents.mlclass': [],
            'dataimport.repository': ['hook_token', 'credentials', 'provider_name', 'git_ref_revisions'],
            'dataimport.worker': [],
            'dataimport.revision': ['message', 'author'],
            'dataimport.workerversion': ['configuration', 'state', 'docker_image', 'docker_image_iid'],
            'images.imageserver': ['s3_bucket', 's3_region', 'created', 'updated', 'validated', 'read_only'],
            'images.image': ['created', 'updated', 'hash', 'status'],
            'documents.element': ['created', 'updated', 'type'],
            'documents.elementpath': [],
            'documents.entity': [],
            'documents.entityrole': [],
            'documents.entitylink': [],
            'documents.transcription': [],
            'documents.transcriptionentity': [],
            'documents.metadata': ['index'],
            'documents.classification': ['ml_class'],
        }

        with open(path, 'r') as file:
            data = json.loads(file.read())

        results = []
        for object in data:
            if object['model'] not in unexpected_fields_by_model:
                continue

            unexpected_fields = unexpected_fields_by_model[object['model']]
            current_fields = list(object['fields'].keys())

            # Delete the id
            del object['pk']

            for field in current_fields:
                # Delete unexpected field
                if field in unexpected_fields:
                    del object['fields'][field]
                # Update corpus field with the new one
                if field == 'corpus' and corpus:
                    object['fields'][field] = corpus

            results.append(object)

        return results

    def test_invalid_file(self):
        with self.assertRaises(CommandError) as context:
            call_command('load_export', 'wrong.txt', '--email', self.user.email)
        self.assertEqual(str(context.exception), 'File wrong.txt is not an SQLite database')

    def test_file_not_exists(self):
        with self.assertRaises(CommandError) as context:
            call_command('load_export', 'wrong.db', '--email', self.user.email)
        self.assertEqual(str(context.exception), 'File wrong.db does not exist')

    def test_invalid_email(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        with self.assertRaises(CommandError) as context:
            call_command('load_export', temp_file, '--email', 'wrong@email')
        self.assertEqual(str(context.exception), 'User with the email wrong@email does not exist')

    def test_invalid_database(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        with self.assertRaises(CommandError) as context:
            call_command('load_export', temp_file, '--email', self.user.email)
        self.assertEqual(str(context.exception), f"The SQLite database {temp_file} is not a correct Arkindex export")

    def test_invalid_version(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        db = sqlite3.connect(temp_file)
        cursor = db.cursor()
        cursor.executescript((BASE_DIR / '../../export/structure.sql').read_text())
        cursor.execute("UPDATE export_version SET version = 1")
        db.commit()
        db.close()

        with self.assertRaises(CommandError) as context:
            call_command('load_export', temp_file, '--email', self.user.email, '--corpus-name', 'My corpus')
        self.assertEqual(str(context.exception), f"The SQLite database {temp_file} does not have the correct export version")

    @patch('arkindex.documents.export.os.unlink')
    @patch('arkindex.project.aws.s3.Object')
    def test_run(self, s3_object_mock, unlink_mock):
        Element.objects.filter(type__folder=False, image__isnull=True).update(
            polygon=LinearRing((0, 0), (0, 1000), (1000, 1000), (1000, 0), (0, 0)),
            image=Image.objects.all().first()
        )
        ImageServer.objects.all().update(validated=True)

        element = self.corpus.elements.get(name='Volume 1')
        transcription = Transcription.objects.first()
        version = WorkerVersion.objects.get(worker__slug='reco')

        element.classifications.create(
            ml_class=self.corpus.ml_classes.create(name='Blah'),
            confidence=.55555555,
        )

        entity1 = self.corpus.entities.create(
            name='Arrokuda',
            type=EntityType.Location,
            metas={'subtype': 'pokemon'},
        )
        entity2 = self.corpus.entities.create(
            name='Stonjourner',
            type=EntityType.Person,
            validated=True,
            moderator=self.superuser,
        )
        role = self.corpus.roles.create(
            parent_name='parent',
            child_name='child',
            parent_type=EntityType.Location,
            child_type=EntityType.Person,
        )
        role.links.create(parent=entity1, child=entity2)

        transcription.transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=version,
        )

        export = self.corpus.exports.create(user=self.user)

        export_corpus(export)

        # Retrieve the database path from the S3 upload argument
        args, _ = s3_object_mock().upload_file.call_args
        db_path = args[0]

        # Call dumpdata command before the deletion
        # Ignore django_rq as it uses a fake database table to insert itself into Django's permissions system
        _, dump_path_before = tempfile.mkstemp(suffix='.json')
        call_command('dumpdata', output=dump_path_before, exclude=['django_rq'])

        # Delete the existing corpus
        corpus_delete(self.corpus.id)
        Image.objects.all().delete()
        ImageServer.objects.all().delete()
        WorkerVersion.objects.filter(id=version.id).delete()

        call_command('load_export', db_path, '--email', self.user.email, '--corpus-name', 'My corpus')

        # Call dumpdata command after the import
        _, dump_path_after = tempfile.mkstemp(suffix='.json')
        call_command('dumpdata', output=dump_path_after, exclude=['django_rq'])

        corpus = Corpus.objects.get(name='My corpus')

        data_before = self.clean_dump_data(dump_path_before, str(corpus.id))
        data_after = self.clean_dump_data(dump_path_after)

        self.assertCountEqual(data_before, data_after)

    def test_run_empty_database(self):
        _, temp_file = tempfile.mkstemp(suffix=".db")
        db = sqlite3.connect(temp_file)
        cursor = db.cursor()
        cursor.executescript((BASE_DIR / '../../export/structure.sql').read_text())
        db.commit()
        db.close()

        call_command('load_export', temp_file, '--email', self.user.email, '--corpus-name', 'My corpus')

        corpus = Corpus.objects.get(name='My corpus')
        self.assertEqual(corpus.types.all().count(), 0)
        self.assertEqual(corpus.ml_classes.all().count(), 0)
        self.assertEqual(corpus.elements.all().count(), 0)
        self.assertEqual(corpus.entities.all().count(), 0)
        self.assertEqual(corpus.roles.all().count(), 0)

    @patch('arkindex.documents.export.os.unlink')
    @patch('arkindex.project.aws.s3.Object')
    def test_run_objects_already_exist(self, s3_object_mock, unlink_mock):
        Element.objects.filter(type__folder=False, image__isnull=True).update(
            polygon=LinearRing((0, 0), (0, 1000), (1000, 1000), (1000, 0), (0, 0)),
            image=Image.objects.all().first()
        )
        ImageServer.objects.all().update(validated=True)

        element = self.corpus.elements.get(name='Volume 1')
        transcription = Transcription.objects.first()
        version = WorkerVersion.objects.get(worker__slug='reco')

        element.classifications.create(
            ml_class=self.corpus.ml_classes.create(name='Blah'),
            confidence=.55555555,
        )

        entity1 = self.corpus.entities.create(
            name='Arrokuda',
            type=EntityType.Location,
            metas={'subtype': 'pokemon'},
        )
        entity2 = self.corpus.entities.create(
            name='Stonjourner',
            type=EntityType.Person,
            validated=True,
            moderator=self.superuser,
        )
        role = self.corpus.roles.create(
            parent_name='parent',
            child_name='child',
            parent_type=EntityType.Location,
            child_type=EntityType.Person,
        )
        role.links.create(parent=entity1, child=entity2)

        transcription.transcription_entities.create(
            entity=entity1,
            offset=1,
            length=1,
            worker_version=version,
        )

        export = self.corpus.exports.create(user=self.user)

        export_corpus(export)

        # Retrieve the database path from the S3 upload argument
        args, _ = s3_object_mock().upload_file.call_args
        db_path = args[0]

        call_command('load_export', db_path, '--email', self.user.email, '--corpus-name', 'My corpus')

        corpus = Corpus.objects.get(name='My corpus')
        self.assertEqual(corpus.types.all().count(), 6)
        self.assertEqual(corpus.ml_classes.all().count(), 1)
        self.assertEqual(corpus.elements.all().count(), 0)
        self.assertEqual(corpus.entities.all().count(), 0)
        self.assertEqual(corpus.roles.all().count(), 0)
