import uuid
from datetime import datetime, timedelta, timezone

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Element, MetaType
from arkindex.project.tests import FixtureAPITestCase


class TestChildrenElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.worker_version = WorkerVersion.objects.get(worker__slug='dla')

    def test_element_children(self):
        lonely_element = Element.objects.get(name='Volume 2, page 2r')
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(lonely_element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page'},
        )

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=surface'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F'
            ],
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?recursive')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 20)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page', 'surface', 'text_line', 'word'},
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?type=page')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r'
            ]
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?name=2')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Act 2',
                'Volume 1, page 2r',
            ]
        )

        nested_volume = self.corpus.elements.create(
            type=self.vol.type,
            name='Nested volume',
        )
        nested_volume.add_parent(self.vol)

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": None,
                "zone": None,
                "rotation_angle": 0,
                "mirrored": False,
                "created": nested_volume.created.isoformat().replace('+00:00', 'Z'),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

        nested_volume.worker_version = self.worker_version
        nested_volume.save()
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": str(self.worker_version.id),
                "zone": None,
                "rotation_angle": 0,
                "mirrored": False,
                "created": nested_volume.created.isoformat().replace('+00:00', 'Z'),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

    def test_element_children_with_zone_true(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_zone=true&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "zone": {
                    "id": str(lonely_element.id),
                    "image" : {
                        "id": str(lonely_element.image.id),
                        "height": lonely_element.image.height,
                        "width": lonely_element.image.width,
                        "path": lonely_element.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.image.server.url,
                        },
                        "status": lonely_element.image.status.value,
                        "url": lonely_element.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace('+00:00', 'Z'),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_with_zone_false(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_zone=false&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace('+00:00', 'Z'),
                "classes": None,
                "has_children": None,
                "thumbnail_url": lonely_element.iiif_thumbnail_url,
            }
        ])

    def test_element_children_with_corpus_false(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_corpus=false&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "zone": {
                    "id": str(lonely_element.id),
                    "image" : {
                        "id": str(lonely_element.image.id),
                        "height": lonely_element.image.height,
                        "width": lonely_element.image.width,
                        "path": lonely_element.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.image.server.url,
                        },
                        "status": lonely_element.image.status.value,
                        "url": lonely_element.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace('+00:00', 'Z'),
                "classes": None,
                "has_children": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_worker_version(self):
        self.corpus.elements.filter(name__contains='page 1r').update(worker_version=self.worker_version)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': str(self.worker_version.id)}
            )

        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
            ]
        )

    def test_element_children_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_element_children_ordering(self):
        expected = [
            'Act 4',
            'Act 5',
            'Act 1',
            'Act 2',
            'Act 3'
        ]
        for ordering, name in enumerate(expected):
            Element.objects.get(type__slug='act', name=name).paths.update(ordering=ordering)

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=act'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            expected,
        )

    def test_children_with_has_children(self):
        # Create a childless element
        element = self.corpus.elements.create(
            name='Spearow',
            type=self.corpus.types.get(slug='surface')
        )
        element.add_parent(self.vol)

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'with_has_children': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['has_children'] for element in response.json()['results']},
            {
                'Act 1': True,
                'Act 2': True,
                'Act 3': True,
                'Act 4': True,
                'Act 5': True,
                'Volume 1, page 1r': True,
                'Volume 1, page 1v': True,
                'Volume 1, page 2r': True,
                'Spearow': False,
            }
        )

    def test_children_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_children_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_children_modified_since(self):
        self.corpus.elements.filter(name='Volume 1, page 1r').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in data['results']},
            {'act', 'page'},
        )

    def test_children_filter_metadata_name(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_name': 'folio'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r',
            ]
        )

    def test_children_filter_metadata_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_children_filter_metadata_name_value(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_name': 'folio', 'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1, page 1v']
        )

    def test_children_filter_metadata_operator(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': ['This option is not supported without metadata_value.']
        })

    def test_children_filter_metadata_operator_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_operator': 'gt', 'metadata_value': 4}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_children_filter_metadata_operator_invalid(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_operator': 'aaaaaa'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': [
                'This metadata operator is not supported.',
                'This option is not supported without metadata_value.',
            ]
        })

    def test_children_filter_numeric_metadata_invalid_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_name': 'a', 'metadata_value': 'lol', 'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['Value must be a float when using a numeric comparison operator.'],
        })

    def test_children_filter_numeric_metadata(self):
        element_type = self.corpus.types.first()
        for i in range(1, 6):
            element = self.corpus.elements.create(type=element_type, name=f'Element {i}')
            element.add_parent(self.vol)
            element.metadatas.create(name='number', type=MetaType.Numeric, value=str(i))
            element.metadatas.create(name='number', type=MetaType.Text, value='a')

        # operator, value, expected element names
        parameters = [
            ('lt', 2, ['Element 1']),
            ('lte', 2, ['Element 1', 'Element 2']),
            ('gt', 3, ['Element 4', 'Element 5']),
            ('gte', 3, ['Element 3', 'Element 4', 'Element 5']),
        ]

        for operator, value, expected_elements in parameters:
            with self.subTest(operator=operator, value=value):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                        data={
                            'metadata_name': 'number',
                            'metadata_value': value,
                            'metadata_operator': operator,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_children_filter_metadata_contains(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={
                    'metadata_name': 'folio',
                    'metadata_operator': 'contains',
                    'metadata_value': '1'
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
            ]
        )

    def test_children_filter_rotation_angle(self):
        element_type = self.corpus.types.first()
        not_rotated = self.corpus.elements.create(type=element_type, name='not rotated')
        rotated = self.corpus.elements.create(type=element_type, name='rotated', rotation_angle=180)
        not_rotated.add_parent(self.vol)
        rotated.add_parent(self.vol)

        cases = [
            (0, ['not rotated']),
            (1, []),
            (180, ['rotated']),
            (181, [])
        ]
        for rotation_angle, expected_elements in cases:
            with self.subTest(rotation_angle=rotation_angle):
                # Requests that return no elements only make 2 SQL queries
                num_queries = 4 if len(expected_elements) else 2
                with self.assertNumQueries(num_queries):
                    response = self.client.get(
                        reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                        data={
                            'name': 'rotated',
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_children_filter_rotation_angle_invalid(self):
        cases = [
            ('to the left', "invalid literal for int() with base 10: 'to the left'"),
            (-1, 'A rotation angle must be between 0 and 359 degrees'),
            (360, 'A rotation angle must be between 0 and 359 degrees'),
            (54.2, "invalid literal for int() with base 10: '54.2'"),
            (4.14j, "invalid literal for int() with base 10: '4.14j'"),
        ]
        for rotation_angle, expected_error in cases:
            with self.subTest(rotation_angle=rotation_angle):
                with self.assertNumQueries(1):
                    response = self.client.get(
                        reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                        data={
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {
                    'rotation_angle': [expected_error]
                })

    def test_children_filter_mirrored(self):
        element_type = self.corpus.types.first()
        mirrored = self.corpus.elements.create(type=element_type, name='mirrored', mirrored=True)
        not_mirrored = self.corpus.elements.create(type=element_type, name='not mirrored', mirrored=False)
        mirrored.add_parent(self.vol)
        not_mirrored.add_parent(self.vol)

        cases = [
            (False, ['not mirrored']),
            (True, ['mirrored']),
        ]
        for mirrored, expected_elements in cases:
            with self.subTest(mirrored=mirrored):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                        data={
                            'name': 'mirrored',
                            'mirrored': mirrored,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_children_invalid_sort(self):
        cases = [
            ({'order': 'blah', 'order_direction': 'asc'}, {'order': ['Unknown sorting field']}),
            ({'order': 'name', 'order_direction': 'left'}, {'order_direction': ['Unknown sorting direction']}),
            ({'order': 'blah', 'order_direction': 'left'}, {'order': ['Unknown sorting field']}),
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.get(
                    reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                    params,
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    def test_children_sort_created(self):
        elements = list(Element.objects.get_descending(self.vol.id).order_by('?').only('id'))
        created = datetime(2020, 2, 2, tzinfo=timezone.utc)
        for element in elements:
            element.created = created
            created += timedelta(seconds=1)
        Element.objects.bulk_update(elements, ['created'])

        cases = [
            ('asc', elements[:20]),
            ('desc', reversed(elements[-20:])),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                    {'order': 'created', 'order_direction': direction, 'recursive': True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['id'] for element in response.json()['results']],
                    [str(element.id) for element in expected]
                )

    def test_children_sort_name(self):
        names = list(Element.objects.get_descending(self.vol.id).order_by('name', 'id').values_list('name', flat=True))

        # Ensure the default position ordering returns a different ordering than by name;
        # otherwise, this test would be useless
        self.assertNotEqual(
            list(Element.objects.get_descending(self.vol.id).values_list('name', flat=True)),
            names,
        )

        cases = [
            ('asc', names[:20]),
            ('desc', list(reversed(names[-20:]))),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                    {'order': 'name', 'order_direction': direction, 'recursive': True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected
                )

    def test_children_sort_position(self):
        ids = list(map(
            str,
            Element
            .objects
            .get_descending(self.vol.id)
            .order_by('paths__ordering', 'id')
            .values_list('id', flat=True)
        ))

        cases = [
            ('asc', ids[:20]),
            ('desc', list(reversed(ids[-20:]))),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse('api:elements-children', kwargs={'pk': self.vol.id}),
                    {'order': 'position', 'order_direction': direction, 'recursive': True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['id'] for element in response.json()['results']],
                    expected
                )
