from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, TextOrientation, Transcription
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.private_corpus.memberships.create(user=cls.user, level=Role.Guest.value)

        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def test_create_transcription_require_login(self):
        response = self.client.post(reverse('api:transcription-create', kwargs={'pk': self.line.id}), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Authentication credentials were not provided.'
        })

    def test_create_transcription_write_right(self):
        self.client.force_login(self.private_read_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
            format='json',
            data={'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_transcription_no_element(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': uuid4()}),
            format='json',
            data={'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_manual_transcription(self):
        """
        Checks the view creates a manual transcription and runs ES indexing
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'text': 'A perfect day in a perfect place'}
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': None,
            'score': None,
            'text': 'A perfect day in a perfect place',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': None,
        })

        new_ts = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertIsNone(new_ts.confidence)
        self.assertEqual(new_ts.worker_version, None)
        self.assertTrue(self.line.transcriptions.filter(pk=new_ts.id).exists())

    def test_create_duplicated_transcription(self):
        """
        Check the view creates a new manual transcriptions with a similar text and element
        """
        self.client.force_login(self.user)
        ts = self.line.transcriptions.create(text='GLOUBIBOULGA')
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'text': ts.text,
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            Transcription.objects.filter(element=self.line, text=ts.text, worker_version__isnull=True).count(),
            2
        )

    def test_create_transcription_with_orientation(self):
        """
        Check that a transcription is created with the specified orientation
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'text': 'A perfect day in a perfect place', 'orientation': 'vertical-lr'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': None,
            'score': None,
            'text': 'A perfect day in a perfect place',
            'orientation': 'vertical-lr',
            'worker_version_id': None,
        })
        new_ts = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertEqual(new_ts.orientation, TextOrientation.VerticalLeftToRight)

    def test_create_transcription_invalid_orientation(self):
        """
        Specifying an invalid text-orientation causes an error
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'text': 'A perfect day in a perfect place', 'orientation': 'wiggly'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_transcription_no_search(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'text': 'A classy text line'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_transcription_worker_version_non_internal(self):
        """
        An internal user is required to create a transcription with a worker version
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'NEKUDOTAYIM',
                'worker_version': str(self.worker_version.id),
                'confidence': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': [
                'An internal user is required to create a transcription refering to a worker_version'
            ]
        })

    def test_create_transcription_worker_version(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'NEKUDOTAYIM',
                'worker_version': str(self.worker_version.id),
                'confidence': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_version, self.worker_version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': .42,
            'score': .42,
            'text': 'NEKUDOTAYIM',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(self.worker_version.id),
        })

    def test_create_transcription_worker_version_score(self):
        """
        Test the deprecated score field is still supported
        """
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'NEKUDOTAYIM',
                'worker_version': str(self.worker_version.id),
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_version, self.worker_version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'confidence': .42,
            'score': .42,
            'text': 'NEKUDOTAYIM',
            'orientation': TextOrientation.HorizontalLeftToRight.value,
            'worker_version_id': str(self.worker_version.id),
        })

    def test_ml_trancription_required_confidence(self):
        """
        A confidence is required when creating a transcription on an element with an worker version
        """
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'CIRCLE',
                'worker_version': str(self.worker_version.id)
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': [
            'Either the score or confidence field should be defined for a transcription with a worker version.'
        ]})

    def test_ml_trancription_score_xor_confidence(self):
        """
        Cannot set both the score and confidence on a transcription with a worker version
        """
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'CIRCLE',
                'score': .42,
                'confidence': .42,
                'worker_version': str(self.worker_version.id)
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': [
            'Either the score or confidence field should be defined for a transcription with a worker version.'
        ]})

    def test_ml_trancription_requires_element_zone(self):
        """
        A worker cannot create a transcription on an element without a zone
        """
        null_zone_page = self.corpus.elements.create(type=self.page.type)

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'text': 'ELLIPSE',
                'confidence': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['The element has no zone']})

    def test_ml_trancription_ignore_type(self):
        """
        An old transcription type can still be sent but will be ignored
        """
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'text': 'CIRCLE',
                'type': 'paragraph',
                'confidence': 0.654816358,
                'worker_version': str(self.worker_version.id)
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        transcription = self.line.transcriptions.get(text='CIRCLE')
        self.assertEqual(transcription.confidence, 0.654816358)
        self.assertEqual(transcription.worker_version, self.worker_version)
