import uuid

from django.contrib.gis.geos import LinearRing
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import (
    Corpus,
    Element,
    Entity,
    EntityLink,
    EntityRole,
    EntityType,
    MetaType,
    TranscriptionEntity,
)
from arkindex.project.tests import FixtureAPITestCase


class TestEntitiesAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.worker_version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_version_2 = WorkerVersion.objects.get(worker__slug='dla')
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.element_type = cls.corpus.types.get(slug='text_line')

    def setUp(self):
        super().setUp()
        self.entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='entity 1',
            worker_version=self.worker_version_1,
        )
        self.entity_bis = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name='entity 2',
            worker_version=self.worker_version_2,
        )
        self.role = EntityRole.objects.create(
            parent_name="parent",
            child_name="child",
            parent_type=EntityType.Person,
            child_type=EntityType.Location,
            corpus=self.corpus
        )
        self.element = self.corpus.elements.create(
            type=self.element_type,
            name='Transcription',
            image=self.page.image,
            polygon=LinearRing((0, 0), (0, 42), (42, 42), (42, 0), (0, 0)),
            worker_version=self.worker_version_1,
        )
        self.transcription = self.element.transcriptions.create(
            text='Some transcribed text',
            worker_version=self.worker_version_1,
        )
        self.metadata = self.element.metadatas.create(
            name='test 1',
            type=MetaType.Text,
            value='Blah',
            element=self.element,
            entity=self.entity,
        )
        self.transcriptionentity = TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity_bis,
            offset=2,
            length=len(self.entity_bis.name)
        )
        self.tr_entities_sample = {
            'entity': str(self.entity.id),
            'offset': 4,
            'length': len(self.entity.name)
        }
        self.tr_entities_confidence_sample = {
            'entity': str(self.entity.id),
            'offset': 4,
            'length': len(self.entity.name),
            'confidence': 1.0
        }
        self.tr_entities_version_sample = {
            'entity': str(self.entity.id),
            'offset': 4,
            'length': len(self.entity.name),
            'worker_version_id': str(self.worker_version_1.id)
        }

    def test_get_entity(self):
        response = self.client.get(reverse('api:entity-details', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.entity.id))
        self.assertEqual(data['type'], self.entity.type.value)
        self.assertEqual(data['corpus']['id'], str(self.corpus.id))
        self.assertEqual(data['name'], self.entity.name)

    def test_get_entity_elements(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.element.id))
        self.assertEqual(results[0]['name'], self.element.name)

    def test_get_entity_elements_corpus_acl(self):
        self.client.force_login(self.user)
        self.element.corpus = self.private_corpus
        self.element.save()
        response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json().get('results'), [])

    def test_get_entity_elements_from_transcription(self):
        elt = self.corpus.elements.create(
            type=self.corpus.types.get(slug='act'),
            name='001',
            image=self.imgsrv.images.first(),
            polygon=LinearRing((0, 0), (0, 200), (200, 200), (200, 0), (0, 0)),
        )
        elt_tr = elt.transcriptions.create(worker_version=self.worker_version_1, text='goodbye')
        TranscriptionEntity.objects.create(transcription=elt_tr, entity=self.entity, offset=42, length=7)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 2)
        self.assertListEqual(
            results,
            [{
                'id': str(e.id),
                'type': e.type.slug,
                'name': e.name,
                'corpus': {
                    'id': str(self.corpus.id),
                    'name': self.corpus.name,
                    'public': self.corpus.public
                },
                'thumbnail_url': None,
                'thumbnail_put_url': None,
                'zone': {
                    'id': str(e.id),
                    'url': e.iiif_url,
                    'polygon': [
                        list(map(int, point))
                        for point in e.polygon.coords
                    ],
                    'image': {
                        'id': str(e.image.id),
                        'width': e.image.width,
                        'height': e.image.height,
                        'path': e.image.path,
                        'url': e.image.url,
                        's3_url': None,
                        'status': e.image.status.value,
                        'server': {
                            'display_name': 'Test Server',
                            'max_height': None,
                            'max_width': None,
                            'url': 'http://server',
                        }
                    }
                },
                'rotation_angle': 0,
                'mirrored': False,
            } for e in (elt, self.element)]
        )

    def test_get_role_in_corpus(self):
        response = self.client.get(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['parent_type'], self.role.parent_type.value)
        self.assertEqual(results[0]['child_type'], self.role.child_type.value)
        self.assertEqual(results[0]['parent_name'], self.role.parent_name)
        self.assertEqual(results[0]['child_name'], self.role.child_name)

    def test_create_role_in_corpus(self):
        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Person.value,
            'child_type': EntityType.Location.value
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        role = self.corpus.roles.get(id=data['id'])
        self.assertEqual(role.parent_type, EntityType.Person)
        self.assertEqual(role.child_type, EntityType.Location)
        self.assertEqual(role.parent_name, 'other parent')
        self.assertEqual(role.child_name, 'other child')

    def test_create_role_already_exist(self):
        data = {
            'parent_name': self.role.parent_name,
            'child_name': self.role.child_name,
            'parent_type': self.role.parent_type.value,
            'child_type': self.role.child_type.value,
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {
            'corpus': ['Role already exists in this corpus'],
            'id': str(self.corpus.id)
        })

    def test_create_role_requires_login(self):
        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Organization.value,
            'child_type': EntityType.Location.value
        }
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_role_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Organization.value,
            'child_type': EntityType.Location.value
        }
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_role_no_corpus_rights(self):
        self.client.force_login(self.user)

        private_corpus = Corpus.objects.create(name='private')
        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Organization.value,
            'child_type': EntityType.Location.value
        }
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(private_corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {
            'corpus': ['You do not have write access to this corpus'],
            'id': [str(private_corpus.id)]
        })

    def test_create_entity_person(self):
        data = {
            'name': 'entity',
            'type': EntityType.Person.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'worker_version': str(self.worker_version_1.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, 'entity')
        self.assertEqual(entity.raw_dates, None)
        self.assertEqual(entity.worker_version, self.worker_version_1)

    def test_create_entity_number(self):
        data = {
            'name': '300g',
            'type': EntityType.Number.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'worker_version': str(self.worker_version_1.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, '300g')
        self.assertEqual(entity.raw_dates, None)
        self.assertEqual(entity.worker_version, self.worker_version_1)

    def test_create_entity_date(self):
        data = {
            'name': '1789',
            'type': EntityType.Date.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'worker_version': str(self.worker_version_1.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, '1789')
        self.assertEqual(entity.raw_dates, entity.name)
        self.assertEqual(entity.worker_version, self.worker_version_1)

    def test_create_entity_requires_login(self):
        data = {
            'name': 'entity',
            'type': EntityType.Person.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'worker_version': str(self.worker_version_1.id)
        }
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_entity_with_worker_version(self):
        data = {
            'name': '1789',
            'type': EntityType.Date.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'worker_version': str(self.worker_version_1.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, '1789')
        self.assertEqual(entity.worker_version, self.worker_version_1)

    def test_create_link(self):
        child = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name="child",
            worker_version=self.worker_version_1,
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-link-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        link = EntityLink.objects.get(id=response.json()['id'])
        self.assertEqual(link.parent.id, self.entity.id)
        self.assertEqual(link.child.id, child.id)
        self.assertEqual(link.role.id, self.role.id)

    def test_create_link_requires_login(self):
        child = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name="child",
            worker_version=self.worker_version_1,
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_link_error(self):
        child = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name="child",
            worker_version=self.worker_version_1,
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-link-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_transcription_entity(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_sample['offset'],
            length=self.tr_entities_sample['length']
        )
        self.assertIsNotNone(transcription_entity)
        self.assertDictEqual(
            response.json(),
            {
                'entity': str(transcription_entity.entity.id),
                'offset': transcription_entity.offset,
                'length': transcription_entity.length,
                'worker_version_id': None,
                'confidence': None
            }
        )

    def test_create_transcription_entity_with_confidence(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_confidence_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_confidence_sample['offset'],
            length=self.tr_entities_confidence_sample['length'],
            confidence=self.tr_entities_confidence_sample['confidence']
        )
        self.assertIsNotNone(transcription_entity)
        self.assertDictEqual(
            response.json(),
            {
                'entity': str(transcription_entity.entity.id),
                'offset': transcription_entity.offset,
                'length': transcription_entity.length,
                'worker_version_id': None,
                'confidence': transcription_entity.confidence
            }
        )

    def test_create_transcription_entity_requires_login(self):
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_transcription_entity_worker_version(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_version_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_sample['offset'],
            length=self.tr_entities_sample['length']
        )
        self.assertEqual(transcription_entity.worker_version, self.worker_version_1)

    def test_create_transcription_entity_worker_version_requires_internal(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_version_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_version_id": ['An internal user is required to link an entity to a transcription with a worker version.']
        })

    def test_create_transcription_entity_wrong_acl(self):
        self.client.force_login(self.user)
        self.element.corpus = self.private_corpus
        self.element.save()
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'transcription': ['invalid UUID or Corpus write-access is forbidden']}
        )

    def test_create_transcription_entity_wrong_length(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data={
                'entity': str(self.entity.id),
                'offset': 4,
                'length': 0
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'length': ['Ensure this value is greater than or equal to 1.']}
        )

    def test_create_transcription_entity_wrong_high_confidence(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data={
                'entity': str(self.entity.id),
                'offset': 4,
                'length': 1,
                'confidence': 2.0
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'confidence': ['Ensure this value is less than or equal to 1.']}
        )

    def test_create_transcription_entity_wrong_low_confidence(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data={
                'entity': str(self.entity.id),
                'offset': 4,
                'length': 1,
                'confidence': -0.5
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'confidence': ['Ensure this value is greater than or equal to 0.']}
        )

    def test_create_transcription_entity_different_corpus(self):
        self.client.force_login(self.user)
        ent = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.private_corpus,
            name="a private entity",
            worker_version=self.worker_version_1,
        )
        self.tr_entities_sample.update({'entity': ent.id})
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'entity': [f'Invalid pk "{ent.id}" - object does not exist.']}
        )

    def test_create_transcription_entity_exists(self):
        self.client.force_login(self.user)
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_transcription_entity_key_missing(self):
        self.client.force_login(self.user)
        data = self.tr_entities_sample
        del data['length']
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=data,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_transcription_entities_open(self):
        response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_transcription_entities_wrong_acl(self):
        self.transcription.element.corpus = self.private_corpus
        self.transcription.element.save()
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 0)

    def test_list_transcription_entities(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertListEqual(
            data['results'],
            [{
                'entity': {
                    'id': str(self.entity_bis.id),
                    'name': self.entity_bis.name,
                    'type': self.entity_bis.type.value,
                    'metas': None,
                    'validated': self.entity_bis.validated,
                    'dates': [],
                    'worker_version_id': str(self.worker_version_2.id),
                },
                'length': self.transcriptionentity.length,
                'offset': self.transcriptionentity.offset,
                'worker_version_id': None,
                'confidence': None
            }]
        )

    def test_list_transcription_entities_worker_version_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}),
                data={'worker_version': 'blah'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}),
                data={'worker_version': str(uuid.uuid4())}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_list_transcription_entities_version_filter(self):
        self.transcriptionentity = TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=8,
            length=len(self.entity_bis.name)
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}),
                {'worker_version': str(self.worker_version_1.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()['results'],
            [{
                'entity': {
                    'id': str(self.entity.id),
                    'name': 'entity 1',
                    'type': 'person',
                    'metas': None,
                    'validated': False,
                    'dates': [],
                    'worker_version_id': str(self.worker_version_1.id),
                },
                'length': 8,
                'offset': 8,
                'worker_version_id': None,
                'confidence': None
            }]
        )

    def test_list_corpus_entities(self):
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-entities', kwargs={'pk': str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            'count': 2,
            'previous': None,
            'next': None,
            'number': 1,
            'results': [
                {
                    'id': str(self.entity.id),
                    'name': 'entity 1',
                    'type': 'person',
                    'validated': False,
                    'dates': [],
                    'metas': None,
                    'worker_version_id': str(self.worker_version_1.id),
                },
                {
                    'id': str(self.entity_bis.id),
                    'name': 'entity 2',
                    'type': 'location',
                    'validated': False,
                    'dates': [],
                    'metas': None,
                    'worker_version_id': str(self.worker_version_2.id),
                }
            ]
        })

    def test_list_corpus_entities_acl(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:corpus-entities', kwargs={'pk': str(self.private_corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_corpus_entities_parent(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-entities', kwargs={'pk': str(self.corpus.id)}),
                {'parent': str(self.element.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'previous': None,
            'next': None,
            'number': 1,
            'results': [
                {
                    'id': str(self.entity_bis.id),
                    'name': 'entity 2',
                    'type': 'location',
                    'validated': False,
                    'dates': [],
                    'metas': None,
                    'worker_version_id': str(self.worker_version_2.id),
                },
            ]
        })

    def test_list_corpus_entities_invalid_parent_uuid(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        with self.assertNumQueries(1):
            response = self.client.get(reverse('api:corpus-entities', kwargs={'pk': str(self.corpus.id)}), {'parent': 'a'})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'parent': ['Parent ID is not a valid UUID']
        })

    def test_list_corpus_entities_parent_not_found(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-entities', kwargs={'pk': str(self.corpus.id)}),
                {'parent': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'},
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_element_entities(self):
        """
        List an element's entities sorted depending on their origin (metadata or transcriptions)
        """
        md = self.element.metadatas.create(name='some_metadata', type=MetaType.Location, value='something')
        md.entity = self.entity_bis
        md.save()
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:element-entities', kwargs={'pk': str(self.element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(
            data,
            {
                'transcriptions': [{
                    'entity': {
                        'id': str(t.entity.id),
                        'name': t.entity.name,
                        'type': t.entity.type.value,
                        'metas': None,
                        'validated': t.entity.validated,
                        'dates': [],
                        'worker_version_id': str(self.worker_version_2.id),
                    },
                    'offset': t.offset,
                    'length': t.length,
                    'worker_version_id': None,
                    'confidence': None
                } for t in TranscriptionEntity.objects.filter(transcription=self.transcription).order_by('offset')],
                'metadata': [{
                    'entity': {
                        'id': str(m.entity.id),
                        'name': m.entity.name,
                        'type': m.entity.type.value,
                        'metas': None,
                        'validated': m.entity.validated,
                        'dates': [],
                        'worker_version_id': str(m.entity.worker_version_id),
                    },
                    'id': str(m.id),
                    'type': m.type.value,
                    'name': m.name,
                    'value': m.value,
                    'dates': []
                } for m in self.element.metadatas.exclude(entity=None)]
            }
        )

    def test_list_element_entities_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:element-entities', kwargs={'pk': str(self.element.id)}),
                data={'worker_version': 'blah'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:element-entities', kwargs={'pk': str(self.element.id)}),
                data={'worker_version': str(uuid.uuid4())}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_list_element_entities_worker_version(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:element-entities', kwargs={'pk': str(self.element.id)}),
                data={'worker_version': str(self.worker_version_1.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            response.json(),
            {
                'transcriptions': [],
                'metadata': [
                    {
                        'id': str(self.metadata.id),
                        'name': 'test 1',
                        'type': 'text',
                        'value': 'Blah',
                        'dates': [],
                        'entity': {
                            'id': str(self.entity.id),
                            'name': 'entity 1',
                            'type': 'person',
                            'validated': False,
                            'dates': [],
                            'metas': None,
                            'worker_version_id': str(self.worker_version_1.id),
                        },
                    }
                ],
            }
        )

    def test_list_element_entities_wrong_acl(self):
        self.element.corpus = self.private_corpus
        self.element.save()
        response = self.client.get(reverse('api:element-entities', kwargs={'pk': str(self.element.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_entity(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:entity-details', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Entity.DoesNotExist):
            Entity.objects.get(id=self.entity.id)

    def test_delete_entity_requires_login(self):
        response = self.client.delete(reverse('api:entity-details', kwargs={'pk': str(self.entity_bis.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:entity-details', kwargs={'pk': str(self.entity_bis.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_validated_entity(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': True},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, True)
        self.assertEqual(self.entity_bis.moderator, self.user)

    def test_unvalidated_entity(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': False},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, False)
        self.assertEqual(self.entity_bis.moderator, None)

    def test_validated_entity_requires_login(self):
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': True},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_unvalidated_entity_requires_login(self):
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': False},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_validated_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': True},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_unvalidated_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': False},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_element_links_not_guest(self):
        """
        A guest access on the element is required to list entity links
        """
        private_elt = self.private_corpus.elements.create(
            type=self.private_corpus.types.create(slug='page'),
            name='A'
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-links', kwargs={'pk': str(private_elt.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have access to this element.'}
        )

    def test_list_element_links(self):
        link = EntityLink.objects.create(parent=self.entity, child=self.entity_bis, role=self.role)
        with self.assertExactQueries('element_links.sql', params={'element_id': self.element.id}):
            response = self.client.get(reverse('api:element-links', kwargs={'pk': str(self.element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertListEqual(
            data['results'],
            [{
                'id': str(link.id),
                'parent': {
                    'id': str(self.entity.id),
                    'name': self.entity.name,
                    'type': self.entity.type.value,
                    'metas': None,
                    'validated': self.entity.validated,
                    'dates': [],
                    'worker_version_id': str(self.worker_version_1.id),
                },
                'child': {
                    'id': str(self.entity_bis.id),
                    'name': self.entity_bis.name,
                    'type': self.entity_bis.type.value,
                    'metas': None,
                    'validated': self.entity_bis.validated,
                    'dates': [],
                    'worker_version_id': str(self.worker_version_2.id),
                },
                'role': {
                    'id': self.role.id,
                    'parent_name': self.role.parent_name,
                    'child_name': self.role.child_name,
                    'parent_type': self.role.parent_type.value,
                    'child_type': self.role.child_type.value
                }
            }]
        )

    def test_list_element_links_not_found(self):
        self.assertFalse(Element.objects.filter(id='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa').exists())
        with self.assertExactQueries('element_links_not_found.sql', params={'element_id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}):
            response = self.client.get(reverse('api:element-links', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
