from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, MLClass
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestRetrieveElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def setUp(self):
        super().setUp()
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')

    def test_get_element(self):
        ml_class = MLClass.objects.create(name='text', corpus=self.corpus)
        classification = self.vol.classifications.create(worker_version=self.worker_version, ml_class=ml_class)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'id': str(self.vol.id),
            'name': 'Volume 1',
            'type': 'volume',
            'corpus': {
                'id': str(self.corpus.id),
                'name': 'Unit Tests',
                'public': True,
            },
            'thumbnail_url': self.vol.thumbnail.s3_url,
            'thumbnail_put_url': None,
            'worker_version': None,
            'zone': None,
            'rotation_angle': 0,
            'mirrored': False,
            'created': '2020-02-02T01:23:45.678000Z',
            'creator': None,
            'rights': ['read'],
            'metadata_count': 0,
            'classifications': [
                {
                    'id': str(classification.id),
                    'confidence': None,
                    'high_confidence': False,
                    'state': 'pending',
                    'worker_version': str(self.worker_version.id),
                    'ml_class': {
                        'id': str(ml_class.id),
                        'name': 'text',
                    }
                }
            ]
        })

    def test_get_element_zone(self):
        self.assertIsNotNone(self.page.image)
        self.assertIsNotNone(self.page.polygon)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))

        self.assertDictEqual(response.json()['zone'], {
            'id': str(self.page.id),
            'image': {
                'id': str(self.page.image.id),
                'path': 'img1',
                'url': 'http://server/img1',
                's3_url': None,
                'width': 1000,
                'height': 1000,
                'status': 'unchecked',
                'server': {
                    'display_name': 'Test Server',
                    'max_width': None,
                    'max_height': None,
                    'url': 'http://server'
                }
            },
            'polygon': [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
            'url': 'http://server/img1/0,0,1000,1000/full/0/default.jpg'
        })

    def test_get_element_thumbnail_folder(self):
        """
        Check getting an element only gives a thumbnail URL with folders
        """
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['thumbnail_url'])
        self.assertIsNone(data['thumbnail_put_url'])

    def test_get_element_thumbnail_put_internal(self):
        """
        Check getting an element only gives a thumbnail PUT URL to bots for folders
        """
        self.client.force_login(self.internal_user)

        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['thumbnail_put_url'], self.vol.thumbnail.s3_put_url)

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()['thumbnail_put_url'])

    def test_get_element_creator(self):
        self.vol.creator = self.user
        self.vol.save()
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['creator'], 'Test user')

    def test_get_element_rights(self):
        cases = [
            (Role.Guest, self.superuser, ['read']),
            (Role.Guest, self.user, ['read']),
            (Role.Contributor, self.superuser, ['read', 'write']),
            (Role.Contributor, self.user, ['read', 'write', 'admin']),
            (Role.Admin, self.superuser, ['read', 'write', 'admin']),
            (Role.Admin, self.user, ['read', 'write', 'admin']),
        ]
        self.client.force_login(self.user)
        for role, creator, expected_rights in cases:
            with self.subTest(role=role, creator=creator):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)
                self.vol.creator = creator
                self.vol.save()
                with self.assertNumQueries(6):
                    response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(response.json()['rights'], expected_rights)
