from django.db import IntegrityError
from django.test import TestCase

from arkindex.images.managers import ImageServerManager
from arkindex.images.models import ImageServer


class TestImageServer(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.root_srv_1 = ImageServer.objects.create(display_name='Root 1', url='http://server.dev/iiif')
        cls.root_srv_2 = ImageServer.objects.create(display_name='Root 2', url='http://server2.dev/iiif/')
        cls.sub_srv_1 = ImageServer.objects.create(display_name='Subfolder 1', url='http://server.dev/iiif/sub')
        cls.sub_srv_2 = ImageServer.objects.create(display_name='Subfolder 2', url='http://server2.dev/iiif/sub')
        cls.empty_srv = ImageServer.objects.create(display_name='Empty', url='http://server.dev/iiif/dev/null')

        cls.root_srv_1.images.create(path='a', width=100, height=100)
        cls.root_srv_1.images.create(path='b', width=100, height=100)
        cls.root_srv_2.images.create(path='sub/a', width=100, height=100)
        cls.root_srv_2.images.create(path='sub/b', width=100, height=100)
        cls.sub_srv_1.images.create(path='c', width=100, height=100)
        cls.sub_srv_1.images.create(path='d', width=100, height=100)
        cls.sub_srv_2.images.create(path='c', width=100, height=100)
        cls.sub_srv_2.images.create(path='d', width=100, height=100)

    def test_merge_empty(self):
        """
        Test merging an empty server into another server does nothing
        """
        self.assertEqual(self.empty_srv.images.count(), 0)
        self.assertEqual(self.root_srv_1.images.count(), 2)

        self.empty_srv.merge_into(self.root_srv_1)

        self.assertEqual(self.empty_srv.images.count(), 0)
        self.assertEqual(self.root_srv_1.images.count(), 2)

    def test_merge_same(self):
        """
        Test merging a server into itself is ignored
        """
        self.assertEqual(self.root_srv_1.images.count(), 2)
        self.root_srv_1.merge_into(self.root_srv_1)
        self.assertEqual(self.root_srv_1.images.count(), 2)

    def test_merge_into_parent(self):
        """
        Test merging a server whose URL is a subfolder of the destination server
        adds the subfolder prefix to each image URL
        Example: /iiif/sub->/iiif will add 'sub' to each image
        """
        self.assertEqual(self.root_srv_1.images.count(), 2)
        self.assertEqual(self.sub_srv_1.images.count(), 2)

        self.sub_srv_1.merge_into(self.root_srv_1)

        self.assertEqual(self.root_srv_1.images.count(), 4)
        self.assertEqual(self.sub_srv_1.images.count(), 0)
        self.assertListEqual(
            list(self.root_srv_1.images.order_by('path').values_list('path', flat=True)),
            ['a', 'b', 'sub/c', 'sub/d'],
        )

    def test_merge_into_child(self):
        """
        Test merging a server A into a server B, when B is a subfolder of A,
        should check that every image is contained into B's subfolder and remove the prefix
        Example: /iiif->/iiif/sub will remove 'sub' from each image
        """
        self.assertEqual(self.root_srv_2.images.count(), 2)
        self.assertEqual(self.sub_srv_2.images.count(), 2)

        self.root_srv_2.merge_into(self.sub_srv_2)

        self.assertEqual(self.root_srv_2.images.count(), 0)
        self.assertEqual(self.sub_srv_2.images.count(), 4)
        self.assertListEqual(
            list(self.sub_srv_2.images.order_by('path').values_list('path', flat=True)),
            ['a', 'b', 'c', 'd'],
        )

    def test_merge_impossible(self):
        """
        Test merging two servers with different subfolders or domains fails
        """
        with self.assertRaises(ValueError):
            self.root_srv_2.merge_into(self.root_srv_1)
        with self.assertRaises(ValueError):
            self.root_srv_1.merge_into(self.root_srv_2)

    def test_merge_into_child_impossible(self):
        """
        When merging a server A into a server B when B is a subfolder of A,
        but not all images are in this subfolder, should raise an exception
        Example: Trying to do /iiif->/iiif/sub except there is a /sub2 image
        """
        sub2_img = self.root_srv_2.images.create(path='sub2/oops', width=100, height=100)
        self.assertEqual(self.root_srv_2.images.count(), 3)
        self.assertEqual(self.sub_srv_2.images.count(), 2)

        with self.assertRaises(ValueError):
            self.root_srv_2.merge_into(self.sub_srv_2)

        sub2_img.delete()

    def test_split(self):
        sub2_img = self.root_srv_2.images.create(path='sub2/splat', width=100, height=100)
        self.assertEqual(ImageServer.objects.count(), 5)
        self.assertEqual(self.root_srv_2.images.count(), 3)

        new_srv = self.root_srv_2.split('sub2')

        self.assertEqual(ImageServer.objects.count(), 6)
        self.assertEqual(self.root_srv_2.images.count(), 2)

        self.assertEqual(new_srv.display_name, 'Root 2_sub2')
        self.assertEqual(new_srv.url, 'http://server2.dev/iiif/sub2')
        self.assertEqual(new_srv.images.count(), 1)
        self.assertEqual(new_srv.images.get(), sub2_img)
        sub2_img.refresh_from_db()
        self.assertEqual(sub2_img.path, 'splat')

        new_srv.delete()
        sub2_img.delete()

    def test_strip_slashes(self):
        """
        Test the server URL is stripped from slashes before saving and the unique constraint gets enforced
        Prevents creating two servers at /iiif and /iiif/
        """
        with self.assertRaisesRegex(IntegrityError, 'unique'):
            ImageServer.objects.create(display_name='Root 1 (bis)', url='http://server.dev/iiif/')

    def test_manager_local(self):
        """
        Test ImageServer has an ImageServerManager which provides the local server as a cached property
        """
        self.assertIsInstance(ImageServer.objects, ImageServerManager)

        # Create a new Manager instance to ensure nothing is already cached
        objects = ImageServerManager()
        objects.model = ImageServer

        with self.settings(LOCAL_IMAGESERVER_ID=self.root_srv_1.id):
            # Ensure only one query is made and the server is cached
            with self.assertNumQueries(1):
                self.assertEqual(objects.local, self.root_srv_1)
                self.assertEqual(objects.local, self.root_srv_1)

    def test_from_url(self):
        """
        Test ImageServer detection from images url
        """
        ImageServer.objects.update(validated=True)

        self.assertEqual(ImageServer.objects.from_url('http://server.dev/iiif/test.jpg'), self.root_srv_1)
        self.assertEqual(ImageServer.objects.from_url('http://server2.dev/iiif/test.jpg'), self.root_srv_2)
        self.assertEqual(ImageServer.objects.count(), 5)

        # A new server will be created
        new_server = ImageServer.objects.from_url('http://new.server.dev/iiif/test.jpg')
        self.assertEqual(ImageServer.objects.count(), 6)
        self.assertEqual(new_server.url, 'http://new.server.dev/iiif')
        self.assertEqual(new_server.display_name, 'http://new.server.dev/iiif')
        self.assertTrue(new_server.validated)

        # And then reused for later calls
        self.assertEqual(ImageServer.objects.from_url('http://new.server.dev/iiif/path%2Fto%2Ffile.jpg'), new_server)
        self.assertEqual(ImageServer.objects.count(), 6)

    def test_from_url_formats(self):
        cases = [
            ('https://whatever.online/prefix/I3f/xxx/test.jpg', 'https://whatever.online/prefix/I3f'),
            ('http://whatever.online/prefixYz/IIIf/path/to/image.jpg', 'http://whatever.online/prefixYz/IIIf'),
            ('https://nubis.univ-paris1.fr/iiif-img/451455/info.json', 'https://nubis.univ-paris1.fr/iiif-img')
        ]
        self.assertEqual(ImageServer.objects.count(), 5)
        for image_url, server_url in cases:
            with self.subTest(image_url=image_url):
                new_server = ImageServer.objects.from_url(image_url)
                # Count incremented: the server just got created
                self.assertEqual(ImageServer.objects.count(), 6)
                self.assertEqual(new_server.url, server_url)
                self.assertEqual(new_server.display_name, server_url)
                self.assertTrue(new_server.validated)
                new_server.delete()
