from io import StringIO

from django.core.management import call_command
from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase


class TestSchemaView(FixtureAPITestCase):

    def test_constant_schema(self):
        """
        Ensure that what a SchemaView outputs is always exactly the same
        as what manage.py spectacular outputs.

        Note that we use the JSON output format instead of the normal YAML because JSON will not allow
        Python-specific objects, while PyYAML would use YAML tags such as "!!python/object",
        and it is much faster to render.
        """
        stdout = StringIO()
        call_command(
            'spectacular',
            format='openapi-json',
            fail_on_warn=True,
            validate=True,
            stdout=stdout,
        )
        expected_schema = stdout.getvalue().strip()

        def _test_schema():
            response = self.client.get(reverse('api:openapi-schema') + '?format=json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.content.decode('utf-8'), expected_schema)

        _test_schema()

        self.client.force_login(self.user)
        _test_schema()

        self.client.force_login(self.superuser)
        _test_schema()

    def test_old_yaml(self):
        """
        Test DRF's `?format=openapi` behaves like Spectacular's `?format=yaml`
        """
        old_response = self.client.get(reverse('api:openapi-schema') + '?format=openapi')
        new_response = self.client.get(reverse('api:openapi-schema') + '?format=yaml')
        self.assertEqual(old_response.status_code, status.HTTP_200_OK)
        self.assertEqual(new_response.status_code, status.HTTP_200_OK)
        self.assertEqual(old_response.content, new_response.content)

    def test_old_json(self):
        """
        Test DRF's `?format=openapi-json` behaves like Spectacular's `?format=json`
        """
        old_response = self.client.get(reverse('api:openapi-schema') + '?format=openapi-json')
        new_response = self.client.get(reverse('api:openapi-schema') + '?format=json')
        self.assertEqual(old_response.status_code, status.HTTP_200_OK)
        self.assertEqual(new_response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(old_response.json(), new_response.json())
